from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING

import lxml.etree
from django.utils.module_loading import import_string

from modernrpc.exceptions import RPCInvalidRequest, RPCMarshallingError, RPCParseError

if TYPE_CHECKING:
    from lxml.etree import _Element

    from modernrpc.types import CustomKwargs
    from modernrpc.xmlrpc.handler import XmlRpcRequest, XmlRpcResult


LXML_DEFAULT_ELEMENT_TYPE = "lxml.etree._Element"


class LxmlDeserializer:
    """xml-rpc deserializer based on the third-party lxml library"""

    def __init__(
        self,
        unmarshaller_klass="modernrpc.xmlrpc.backends.marshalling.EtreeElementUnmarshaller",
        unmarshaller_kwargs: CustomKwargs = None,
        element_type_klass=LXML_DEFAULT_ELEMENT_TYPE,
        load_parser_kwargs: CustomKwargs = None,
        load_kwargs: CustomKwargs = None,
    ):
        self.unmarshaller_klass = import_string(unmarshaller_klass)
        self.unmarshaller_kwargs = unmarshaller_kwargs or {}

        self.element_type_klass = import_string(element_type_klass)

        self.load_parser_kwargs = load_parser_kwargs or {}
        self.load_parser_kwargs.setdefault("resolve_entities", False)  # Prevent entity expansion
        self.load_parser_kwargs.setdefault("no_network", True)  # Prevent network access
        self.load_parser_kwargs.setdefault("dtd_validation", False)  # Disable DTD validation
        self.load_parser_kwargs.setdefault("load_dtd", False)  # Disable DTD loading
        self.load_parser_kwargs.setdefault("huge_tree", False)  # Prevent the billion-laugh attack

        self.load_kwargs = load_kwargs or {}

    @cached_property
    def unmarshaller(self):
        return self.unmarshaller_klass[self.element_type_klass](**self.unmarshaller_kwargs)

    def loads(self, data: str) -> XmlRpcRequest:
        if "parser" not in self.load_kwargs:
            # Create a custom parser, with default secure params, configurable from settings
            parser = lxml.etree.XMLParser(**self.load_parser_kwargs)
            self.load_kwargs["parser"] = parser

        try:
            root_obj: _Element = lxml.etree.fromstring(data, **self.load_kwargs)
        except lxml.etree.XMLSyntaxError as exc:
            raise RPCParseError(str(exc)) from exc

        try:
            return self.unmarshaller.element_to_request(root_obj)
        except Exception as exc:
            raise RPCInvalidRequest(str(exc)) from exc


class LxmlSerializer:
    """xml-rpc serializer based on the third-party lxml library"""

    def __init__(
        self,
        marshaller_klass="modernrpc.xmlrpc.backends.marshalling.EtreeElementMarshaller",
        marshaller_kwargs: CustomKwargs = None,
        element_type_klass=LXML_DEFAULT_ELEMENT_TYPE,
        dump_kwargs: CustomKwargs = None,
    ):
        self.marshaller_klass = import_string(marshaller_klass)
        self.marshaller_kwargs = marshaller_kwargs or {}
        self.marshaller_kwargs.setdefault("element_factory", "lxml.etree.Element")
        self.marshaller_kwargs.setdefault("sub_element_factory", "lxml.etree.SubElement")

        self.element_type_klass = import_string(element_type_klass)

        self.dump_kwargs = dump_kwargs or {}

    @cached_property
    def marshaller(self):
        elt_factory = import_string(self.marshaller_kwargs.pop("element_factory"))
        sub_elt_factory = import_string(self.marshaller_kwargs.pop("sub_element_factory"))
        return self.marshaller_klass[self.element_type_klass](elt_factory, sub_elt_factory, **self.marshaller_kwargs)

    def dumps(self, result: XmlRpcResult) -> str:
        """Serialize an XmlRpcResult to an XML string."""
        try:
            root = self.marshaller.result_to_element(result)
        except Exception as exc:
            raise RPCMarshallingError(result.data, exc) from exc

        return lxml.etree.tostring(root, encoding="unicode", **self.dump_kwargs)



from __future__ import annotations

import importlib
import logging
from collections import OrderedDict, defaultdict
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Callable, Iterable

from asgiref.sync import async_to_sync, iscoroutinefunction, sync_to_async
from django.utils.functional import cached_property

from modernrpc import Protocol
from modernrpc.compat import is_union_type, union_str_repr
from modernrpc.config import settings
from modernrpc.constants import NOT_SET
from modernrpc.exceptions import (
    AuthenticationError,
    RPCInvalidParams,
)
from modernrpc.helpers import check_flags_compatibility, ensure_sequence
from modernrpc.introspection import DocstringParser, Introspector

if TYPE_CHECKING:
    from django.http import HttpRequest

    from modernrpc.handler import RpcHandler
    from modernrpc.server import RpcServer
    from modernrpc.types import AuthPredicateType

logger = logging.getLogger(__name__)


@dataclass
class RpcRequestContext:
    """
    Represents the context of an RPC request.

    This class serves as a container for the components involved in the processing of an RPC request.
    It holds references to the original http request, server, handler, protocol, and optionally, the authentication
    result, providing access to contextual information needed during the RPC's lifecycle.

    :ivar request: The HTTP request associated with the RPC context.
    :ivar server: The RPC server handling the request.
    :ivar handler: The handler responsible for processing the RPC request.
    :ivar protocol: The protocol used for the RPC.
    :ivar auth_result: The result of authentication for this RPC context, if applicable.
    """

    request: HttpRequest
    server: RpcServer
    handler: RpcHandler
    protocol: Protocol
    auth_result: Any = None


@dataclass
class ProcedureArgDocs:
    docstring: str = ""
    doc_type: str = ""
    type_hint: type | None = None

    @property
    def type_hint_as_str(self) -> str:
        if not self.type_hint:
            return ""
        if is_union_type(self.type_hint):
            return union_str_repr(self.type_hint)
        return self.type_hint.__name__

    @property
    def documented_type(self) -> str:
        return self.type_hint_as_str or self.doc_type


class ProcedureWrapper:
    """
    Wraps a python global function to be used to extract information and call the concrete procedure.
    """

    def __init__(
        self,
        func_or_coro: Callable,
        name: str | None = None,
        protocol: Protocol = Protocol.ALL,
        auth: AuthPredicateType = NOT_SET,
        context_target: str | None = None,
    ) -> None:
        # Store the reference to the registered function
        self.func_or_coro = func_or_coro

        # @decorator parameters
        self.name = name or func_or_coro.__name__
        self.protocol = protocol
        self.context_target = context_target

        self.auth = auth

    @cached_property
    def introspector(self) -> Introspector:
        return Introspector(self.func_or_coro)

    @cached_property
    def docstring_parser(self) -> DocstringParser:
        return DocstringParser(self.func_or_coro)

    def __repr__(self) -> str:
        return f'ModernRPC Procedure "{self.name}"'

    def __str__(self) -> str:
        return f"{self.name}({', '.join(self.arguments.keys())})"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ProcedureWrapper):
            return NotImplemented
        return (
            self.func_or_coro == other.func_or_coro
            and self.name == other.name
            and self.protocol == other.protocol
            and self.auth == other.auth
        )

    def __hash__(self):
        return hash(self.func_or_coro)

    def check_permissions(self, request: HttpRequest) -> Any:
        """
        Call each predicate associated with the procedure to check authentication.
        If any of the predicate returns a truthy value, it is returned as result.
        In other cases, raise an AuthenticationFailed exception.
        """
        if self.auth == NOT_SET or not self.auth:
            return True

        for callback in ensure_sequence(self.auth):
            if result := callback(request):
                return result

        raise AuthenticationError(self.name)

    def execute(
        self,
        context: RpcRequestContext,
        args: Iterable[Any] | None = None,
        kwargs: dict | None = None,
    ) -> Any:
        args = args or []
        kwargs = kwargs or {}

        try:
            context.auth_result = self.check_permissions(context.request)
        except Exception as exc:
            raise AuthenticationError(self.name) from exc

        # If the remote procedure requested access to context data, provide it into proper kwargs key
        if self.context_target:
            kwargs[self.context_target] = context

        logger.debug("Params: args = %s - kwargs = %s", args, kwargs)

        function = async_to_sync(self.func_or_coro) if iscoroutinefunction(self.func_or_coro) else self.func_or_coro

        try:
            return function(*args, **kwargs)

        except TypeError as exc:
            # If given params cannot be transmitted properly to the procedure function
            raise RPCInvalidParams(str(exc)) from None

    async def aexecute(
        self,
        context: RpcRequestContext,
        args: Iterable[Any] | None = None,
        kwargs: dict | None = None,
    ) -> Any:
        args = args or []
        kwargs = kwargs or {}

        try:
            auth_result = self.check_permissions(context.request)
        except Exception as exc:
            raise AuthenticationError(self.name) from exc

        # If the remote procedure requested access to context data, provide it into proper kwargs key
        if self.context_target:
            context.auth_result = auth_result
            kwargs[self.context_target] = context

        logger.debug("Params: args = %s - kwargs = %s", args, kwargs)

        coro = self.func_or_coro if iscoroutinefunction(self.func_or_coro) else sync_to_async(self.func_or_coro)

        try:
            return await coro(*args, **kwargs)
        except TypeError as exc:
            # If given params cannot be transmitted properly to the procedure function
            raise RPCInvalidParams(str(exc)) from None

    @cached_property
    def is_available_in_xml_rpc(self):
        return check_flags_compatibility(self.protocol, Protocol.XML_RPC)

    @cached_property
    def is_available_in_json_rpc(self):
        return check_flags_compatibility(self.protocol, Protocol.JSON_RPC)

    @cached_property
    def text_doc(self) -> str:
        """Method docstring, as raw text"""
        return self.docstring_parser.docstring

    @cached_property
    def html_doc(self) -> str:
        """Convert the text part of the docstring to an HTML representation, using the parser set in settings"""
        if not self.text_doc:
            return ""

        if settings.MODERNRPC_DOC_FORMAT.lower() in ("rst", "restructured", "restructuredtext"):
            docutils = importlib.import_module("docutils.core")
            return docutils.publish_parts(self.text_doc, writer_name="html")["body"]

        if settings.MODERNRPC_DOC_FORMAT.lower() in ("md", "markdown"):
            markdown_module = importlib.import_module("markdown")
            return markdown_module.markdown(self.text_doc)

        html_content = self.text_doc.replace("\n\n", "</p><p>").replace("\n", " ")
        return f"<p>{html_content}</p>"

    @cached_property
    def arguments_names(self) -> list[str]:
        """
        List of arguments of the wrapped function.

        If you need to access the documentation for each argument, use the `arguments` property instead.
        """
        return [arg for arg in self.introspector.signature.parameters if arg != self.context_target]

    @cached_property
    def arguments(self) -> OrderedDict[str, ProcedureArgDocs]:
        result: defaultdict[str, ProcedureArgDocs] = defaultdict(ProcedureArgDocs)
        for param in self.arguments_names:
            result[param].doc_type = self.docstring_parser.args_types.get(param, "")
            result[param].docstring = self.docstring_parser.args_docstrings.get(param, "")
            result[param].type_hint = self.introspector.get_arg_type_hint(param)

        return OrderedDict(result)

    @cached_property
    def returns(self) -> ProcedureArgDocs:
        return ProcedureArgDocs(
            docstring=self.docstring_parser.return_doc,
            doc_type=self.docstring_parser.return_type,
            type_hint=self.introspector.get_return_type_hint(),
        )



from __future__ import annotations

from enum import Flag, auto

SYSTEM_NAMESPACE_DOTTED_PATH = "modernrpc.system_procedures.system"


class Protocol(Flag):
    """Define a custom type to use everywhere a protocol (JSON-RPC or XML-RPC) is expected"""

    JSON_RPC = auto()
    XML_RPC = auto()

    ALL = JSON_RPC | XML_RPC


class Id:
    def __init__(self, name: str) -> None:
        self.name = name

    def __repr__(self):
        return self.name


NOT_SET = Id("NOT_SET")



from __future__ import annotations

import base64
import functools
from typing import TYPE_CHECKING, Callable
from urllib.parse import unquote

if TYPE_CHECKING:
    from django.http import HttpRequest


def extract_header(request: HttpRequest, header_name: str) -> str:
    """Extract a header from a request object or raise a ValueError when it is not found"""
    try:
        return request.headers[header_name]
    except Exception as exc:
        raise ValueError(f'Unable to find "{header_name}" header in request') from exc


def extract_generic_token(request: HttpRequest, header_name: str, auth_type: str) -> str:
    """
    Extract a generic token from a request object and raise a ValueError when it is not found

    :param request: HTTP request containing the headers
    :param header_name: Header to lookup for authentication data (ex: Authorization)
    :param auth_type: Type of authentication information to extract (ex: Bearer, Basic, etc.)
    :return: Authentication data
    """
    auth_header = extract_header(request, header_name)
    current_auth_type, credentials = auth_header.split()

    if current_auth_type.lower() != auth_type.lower():
        raise ValueError(f'Invalid authentication type. Expected "{auth_type}", found "{current_auth_type}"')

    return unquote(credentials)


def extract_http_basic_auth(request: HttpRequest) -> tuple[str, str]:
    """Extract HTTP Basic Auth credentials from a request object. Return a tuple with username and password"""
    credentials = extract_generic_token(request, "Authorization", "Basic")

    uname, passwd = base64.b64decode(credentials).decode("utf-8").split(":")
    return unquote(uname), unquote(passwd)


extract_bearer_token: Callable[[HttpRequest], str] = functools.partial(
    extract_generic_token, header_name="Authorization", auth_type="Bearer"
)
extract_bearer_token.__doc__ = """
Extract a Bearer token from a request object. Return the token.

:param request: HTTP request containing the headers
:return: Token string
"""



import sys
import types
import typing

import django
from django.views.decorators.csrf import csrf_exempt

if django.VERSION >= (5, 0):
    # Django updated its decorators to support wrapping asynchronous method in release 5.0
    # REF: https://docs.djangoproject.com/en/5.2/releases/5.0/#decorators

    async_csrf_exempt = csrf_exempt

else:
    # Fore Django version < 5.0, we partially redefine decorators to support async view

    def async_csrf_exempt(view):
        view.csrf_exempt = True
        return view


if sys.version_info >= (3, 14):

    def is_union_type(_type: type):
        return isinstance(_type, typing.Union)

elif sys.version_info >= (3, 10):

    def is_union_type(_type: type):
        return typing.get_origin(_type) in (types.UnionType, typing.Union)

else:

    def is_union_type(_type: type):
        return typing.get_origin(_type) is typing.Union


if sys.version_info >= (3, 14):

    def union_str_repr(obj):
        return str(obj)

else:

    def union_str_repr(obj: type):
        return " | ".join(t.__name__ for t in typing.get_args(obj))



# PEP 585: use of list[Any] instead of List[Any] is available since Python 3.9, enable it for older versions
# PEP 604: use of typeA | typeB is available since Python 3.10, enable it for older versions
from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING, Iterable

import rapidjson
from django.utils.module_loading import import_string
from rapidjson import JSONDecodeError

from modernrpc.exceptions import RPCMarshallingError, RPCParseError

if TYPE_CHECKING:
    from modernrpc.jsonrpc.handler import JsonRpcRequest, JsonRpcResult
    from modernrpc.types import CustomKwargs, DictStrAny


class RapidjsonDeserializer:
    """json-rpc deserializer based on the third-party simplejson library"""

    def __init__(
        self,
        unmarshaller_klass="modernrpc.jsonrpc.backends.marshalling.Unmarshaller",
        unmarshaller_kwargs: CustomKwargs = None,
        load_kwargs: CustomKwargs = None,
    ):
        self.unmarshaller_klass = import_string(unmarshaller_klass)
        self.unmarshaller_kwargs = unmarshaller_kwargs or {}

        self.load_kwargs = load_kwargs or {}

    @cached_property
    def unmarshaller(self):
        return self.unmarshaller_klass(**self.unmarshaller_kwargs)

    def loads(self, data: str) -> JsonRpcRequest | list[JsonRpcRequest]:
        try:
            structured_data: list[DictStrAny] | DictStrAny = rapidjson.loads(data, **self.load_kwargs)
        except JSONDecodeError as exc:
            raise RPCParseError(str(exc), data=exc) from exc

        return self.unmarshaller.dict_to_request(structured_data)


class RapidjsonSerializer:
    """json-rpc serializer based on the third-party simplejson library"""

    def __init__(
        self,
        marshaller_klass="modernrpc.jsonrpc.backends.marshalling.Marshaller",
        marshaller_kwargs: CustomKwargs = None,
        dump_kwargs: CustomKwargs = None,
    ):
        self.marshaller_klass = import_string(marshaller_klass)
        self.marshaller_kwargs = marshaller_kwargs or {}

        self.dump_kwargs = dump_kwargs or {}
        self.dump_kwargs.setdefault("datetime_mode", rapidjson.DM_ISO8601)

    @cached_property
    def marshaller(self):
        return self.marshaller_klass(**self.marshaller_kwargs)

    def dumps(self, result: JsonRpcResult | Iterable[JsonRpcResult]) -> str:
        structured_data = self.marshaller.result_to_dict(result)
        try:
            return rapidjson.dumps(structured_data, **self.dump_kwargs)
        except (TypeError, UnicodeDecodeError) as exc:
            raise RPCMarshallingError(structured_data, exc) from exc



from typing import Any

# Constantes used in builtin exceptions
RPC_PARSE_ERROR = -32700
RPC_INVALID_REQUEST = -32600
RPC_METHOD_NOT_FOUND = -32601
RPC_INVALID_PARAMS = -32602
RPC_INTERNAL_ERROR = -32603

# Used as a minimal value for any custom error returned by the server
RPC_CUSTOM_ERROR_BASE = -32099
# Used as a maximal value for any custom error returned by the server
RPC_CUSTOM_ERROR_MAX = -32000


class RPCException(Exception):
    """
    Base class for all RPC exceptions. Custom exceptions raised by your remote procedures
    should inherit from RPCException.
    """

    def __init__(self, code: int, message: str, data: Any = None):
        super().__init__(f"[{code}] {message}")
        self.code = code
        self.message = message
        self.data = data


class RPCParseError(RPCException):
    """Raised by handlers if the request can't be read as valid JSON or XML data."""

    def __init__(self, message: str, data: Any = None):
        err_msg = f"Parse error, unable to read the request: {message}"
        super().__init__(RPC_PARSE_ERROR, err_msg, data)


class RPCInsecureRequest(RPCParseError):
    """Raised in backends if the request is maliciously formed."""

    def __init__(self, message: str, data: Any = None):
        err_msg = f"Security error: {message}"
        super().__init__(err_msg, data)


class RPCInvalidRequest(RPCException):
    """Raised by handlers if incoming JSON or XML data is not a valid JSON-RPC or XML-RPC data."""

    def __init__(self, message: str, data: Any = None):
        err_msg = f"Invalid request: {message}"
        super().__init__(RPC_INVALID_REQUEST, err_msg, data)


class RPCMethodNotFound(RPCException):
    """Raised by handlers when no procedure was found with the given name on the server (for the current protocol)."""

    def __init__(self, name: str, data: Any = None):
        err_msg = f'Method not found: "{name}"'
        super().__init__(RPC_METHOD_NOT_FOUND, err_msg, data)


class RPCInvalidParams(RPCException):
    """Raised by handlers if the request parameters do not match the procedure's expected ones."""

    def __init__(self, message: str, data: Any = None):
        err_msg = f"Invalid parameters: {message}"
        super().__init__(RPC_INVALID_PARAMS, err_msg, data)


class RPCInternalError(RPCException):
    """Raised by handlers if any standard exception is raised during the execution of the procedure."""

    def __init__(self, message: str, data: Any = None):
        err_msg = f"Internal error: {message}"
        super().__init__(RPC_INTERNAL_ERROR, err_msg, data)


class RPCMarshallingError(RPCException):
    def __init__(self, data: Any, exc: Exception):
        super().__init__(
            RPC_INTERNAL_ERROR,
            f"Unable to serialize result data: {data}. Original exception: {exc}",
        )


class AuthenticationError(RPCException):
    """Raised when the authentication system forbids execution of a remote procedure."""

    def __init__(self, method_name: str):
        super().__init__(
            RPC_INTERNAL_ERROR,
            f'Authentication failed when calling "{method_name}"',
        )


# In 1.0, RPCUnknownMethod were renamed to RPCMethodNotFound
# Set an alias for backward compatibility
RPCUnknownMethod = RPCMethodNotFound



from django.conf import settings as user_settings

from modernrpc.config import default_settings


class ModernRpcSettings:
    def __getattr__(self, item):
        try:
            # First, try to retrieve setting from the project level settings module
            return getattr(user_settings, item)
        except AttributeError:
            # Fallback: return the default value, provided by django-modern-rpc
            return getattr(default_settings, item)


settings = ModernRpcSettings()



from __future__ import annotations

import functools
import logging
from http import HTTPStatus
from typing import TYPE_CHECKING, Any, Callable

from django.http import HttpResponse, HttpResponseNotAllowed
from django.shortcuts import redirect
from django.utils.log import log_response
from django.utils.module_loading import import_string
from django.views.decorators.csrf import csrf_exempt

from modernrpc import Protocol, RpcRequestContext
from modernrpc.compat import async_csrf_exempt
from modernrpc.config import settings
from modernrpc.constants import NOT_SET, SYSTEM_NAMESPACE_DOTTED_PATH
from modernrpc.core import ProcedureWrapper
from modernrpc.exceptions import RPCException, RPCInternalError, RPCMethodNotFound
from modernrpc.helpers import check_flags_compatibility, first_true
from modernrpc.views import handle_rpc_request, handle_rpc_request_async

if TYPE_CHECKING:
    from django.http import HttpRequest
    from django.shortcuts import SupportsGetAbsoluteUrl

    from modernrpc.handler import RpcHandler
    from modernrpc.types import AuthPredicateType


logger = logging.getLogger(__name__)

RpcErrorHandler = Callable[[BaseException, RpcRequestContext], None]


class RegistryMixin:
    def __init__(self, auth: AuthPredicateType = NOT_SET) -> None:
        self._registry: dict[str, ProcedureWrapper] = {}
        self.auth = auth

    def register_procedure(
        self,
        procedure: Callable | None = None,
        name: str | None = None,
        protocol: Protocol = Protocol.ALL,
        auth: AuthPredicateType = NOT_SET,
        context_target: str | None = None,
    ) -> Callable:
        """
        Registers a procedure for handling RPC (Remote Procedure Call) requests. This function can be used as a
        decorator to register the given callable function as an RPC procedure.

        :param procedure: A callable function to be registered as an RPC procedure
        :param name: A custom name for the procedure. If not provided, the callable function's name is used by default.
        :param protocol: Specifies the protocol (e.g., JSON-RPC, XML-RPC) under which the procedure can be executed.
                         Defaults: Protocol.ALL
        :param auth: Defines user authentication settings or access rules for the procedure. Defaults to NOT_SET, in
                     which case the server's authentication settings will be used.
        :param context_target: Specify the procedure argument name for accessing the RPC request context.

        :raises ValueError: If a procedure can't be registered
        """

        def decorated(func: Callable) -> Callable:
            if name and name.startswith("rpc."):
                raise ValueError(
                    'According to JSON-RPC specs, method names starting with "rpc." are reserved for system extensions '
                    "and must not be used. See https://www.jsonrpc.org/specification#extensions for more information."
                )

            auth_predicate = self.auth if auth == NOT_SET else auth
            wrapper = ProcedureWrapper(
                func, name, protocol=protocol, auth=auth_predicate, context_target=context_target
            )

            if wrapper.name in self._registry and wrapper != self._registry[wrapper.name]:
                raise ValueError(f"Procedure {wrapper.name} is already registered")

            self._registry[wrapper.name] = wrapper
            logger.debug(f"Registered procedure {wrapper.name} into {self.__class__.__name__} instance")
            return func

        # If @server.register_procedure() is used with parenthesis (with or without argument)
        if procedure is None:
            return decorated

        # If @server.register_procedure is used without a parenthesis
        return decorated(procedure)

    @property
    def procedures(self) -> dict[str, ProcedureWrapper]:
        return self._registry


class RpcNamespace(RegistryMixin): ...


class RpcServer(RegistryMixin):
    def __init__(
        self,
        register_system_procedures: bool = True,
        supported_protocol: Protocol = Protocol.ALL,
        auth: AuthPredicateType = NOT_SET,
        error_handler: RpcErrorHandler | None = None,
        allow_get_method: str | Callable[..., Any] | SupportsGetAbsoluteUrl | None = None,
        default_encoding: str = settings.MODERNRPC_DEFAULT_ENCODING,
    ) -> None:
        super().__init__(auth)
        self.handler_klasses = list(
            filter(
                lambda cls: check_flags_compatibility(cls.protocol, supported_protocol),
                (import_string(klass) for klass in settings.MODERNRPC_HANDLERS),
            )
        )

        if register_system_procedures:
            system_namespace = import_string(SYSTEM_NAMESPACE_DOTTED_PATH)
            self.register_namespace(system_namespace, "system")

        self.error_handler = error_handler
        self.redirect_to = allow_get_method
        self.default_encoding = default_encoding

    def register_namespace(self, namespace: RpcNamespace, name: str | None = None) -> None:
        if name:
            prefix = name + "."
            logger.debug(
                f"About to register {len(namespace.procedures)} procedure(s) "
                f"from namespace '{name}' into the top-level server"
            )
        else:
            prefix = ""
            logger.debug(
                f"About to register {len(namespace.procedures)} procedure(s) "
                f"from unnamed namespace into the top-level server"
            )

        for procedure_name, wrapper in namespace.procedures.items():
            self.register_procedure(
                wrapper.func_or_coro,
                name=f"{prefix}{procedure_name}",
                protocol=wrapper.protocol,
                auth=wrapper.auth,
                context_target=wrapper.context_target,
            )

    def get_procedure_wrapper(self, name: str, protocol: Protocol) -> ProcedureWrapper:
        """Return the procedure wrapper with given name compatible with given protocol, or raise RPCMethodNotFound"""
        try:
            wrapper = self.procedures[name]
        except KeyError:
            raise RPCMethodNotFound(name) from None

        if check_flags_compatibility(wrapper.protocol, protocol):
            return wrapper

        raise RPCMethodNotFound(name) from None

    def get_request_handler(self, request: HttpRequest) -> RpcHandler | None:
        handler_klass = first_true(self.handler_klasses, pred=lambda cls: cls.can_handle(request))
        try:
            return handler_klass()
        except TypeError:
            # first_true() returned None -> TypeError: 'NoneType' object is not callable
            return None

    def on_error(self, exception: BaseException, context: RpcRequestContext) -> RPCException:
        """
        If an error handler is defined, call it to run arbitrary code and return its not-None result.
        Else, check the given exception type and return it if it is a subclass of RPCException.
        Convert any other exception into RPCInternalError.
        """
        if self.error_handler:
            try:
                self.error_handler(exception, context)
            except Exception as exc:
                exception = exc

        if isinstance(exception, RPCException):
            return exception
        return RPCInternalError(message=str(exception))

    def build_method_not_allowed_reponse(self, request: HttpRequest) -> HttpResponseNotAllowed:
        """
        Build an HttpResponseNotAllowed instance with the correct list of allowed methods.
        """
        allowed_methods = ["GET", "POST"] if self.redirect_to else ["POST"]
        response = HttpResponseNotAllowed(allowed_methods)
        log_response(f"Method Not Allowed ({request.method}): {request.path}", response=response, request=request)
        return response

    def check_request(self, request: HttpRequest) -> HttpResponse | None:
        """
        Check incoming request for common issues. When everything is fine, return None. Else, return the appropriate
        HttpResponse instance.

        :param request: Request instance as received by Django
        :return: A response instance (HttpResponsePermanentRedirect, HttpResponseNotAllowed, HttpResponse) or None
        """
        if request.method == "GET":
            if self.redirect_to:
                return redirect(to=self.redirect_to, permanent=True)

            return self.build_method_not_allowed_reponse(request)

        if request.method != "POST":
            return self.build_method_not_allowed_reponse(request)

        if not request.content_type:
            return HttpResponse(
                "Unable to handle your request, the Content-Type header is mandatory to allow server "
                "to determine which handler can interpret your request.",
                status=HTTPStatus.BAD_REQUEST,
                content_type="text/plain",
            )
        return None

    @staticmethod
    def build_response(handler: RpcHandler, result_data: str | tuple[int, str]) -> HttpResponse:
        if isinstance(result_data, tuple) and len(result_data) == 2:
            status, result_data = result_data
        else:
            status = HTTPStatus.OK

        return HttpResponse(result_data, status=status, content_type=handler.response_content_type)

    @property
    def view(self) -> Callable:
        """
        Returns a synchronous view function that can be used in Django URL patterns.
        The view is decorated with csrf_exempt and require_POST.

        :return: A callable view function
        """
        view_func = functools.partial(handle_rpc_request, server=self)
        return csrf_exempt(view_func)

    @property
    def async_view(self) -> Callable:
        """
        Returns an asynchronous view function that can be used in Django URL patterns.
        The view is decorated with csrf_exempt and require_POST.

        :return: An awaitable async view function
        """
        view_func = functools.partial(handle_rpc_request_async, server=self)
        return async_csrf_exempt(view_func)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Callable, Dict, Generic, Optional, Sequence, TypeVar, Union


@dataclass
class RpcRequest:
    """Basic information about RPC request"""

    method_name: str
    args: list[Any] | tuple[Any] = field(default_factory=list)


RequestType = TypeVar("RequestType", bound=RpcRequest)


@dataclass
class RpcResult(Generic[RequestType]):
    request: RequestType


@dataclass
class RpcSuccessResult(RpcResult[RequestType]):
    data: Any = None


@dataclass
class RpcErrorResult(RpcResult[RequestType]):
    code: int
    message: str
    # This field is only dumped to JSON-RPC clients!
    data: Any = None


# Type aliases.
# Each one may be typed with typing.TypeAlias as soon as Python 3.10 will be the minimal requirement
# New syntax `type X = list[str]` will be available from Python 3.12
DictStrAny = Dict[str, Any]  # Not sure why, but using dict instead of typing.Dict here will cause Python 3.8 to crash
CustomKwargs = Optional[DictStrAny]

NotSetType = object
AuthPredicate = Callable[[RpcRequest], bool]
AuthPredicateType = Union[NotSetType, AuthPredicate, Sequence[AuthPredicate]]



from __future__ import annotations

import datetime
import xmlrpc.client
from typing import TYPE_CHECKING, Any, Iterable, Sequence

from modernrpc.constants import NOT_SET

if TYPE_CHECKING:
    from enum import Flag
    from typing import Callable


def check_flags_compatibility(a: Flag, b: Flag) -> bool:
    """Check that both flags are compatible"""
    return bool(a & b)


def get_builtin_date(
    date: str | datetime.datetime | xmlrpc.client.DateTime,
    date_format: str = "%Y-%m-%dT%H:%M:%S",
    raise_exception: bool = False,
) -> datetime.datetime | None:
    """
    Try to convert a date to a builtin instance of ``datetime.datetime``.
    The input date can be a ``str``, a ``datetime.datetime``, a ``xmlrpc.client.Datetime`` or a ``xmlrpclib.Datetime``
    instance. The returned object is a ``datetime.datetime``.

    :param date: The date object to convert.
    :param date_format: If the given date is a str, it is passed to strptime for parsing
    :param raise_exception: If set to True, an exception will be raised if the input string cannot be parsed
    :return: A valid ``datetime.datetime`` instance
    """
    if isinstance(date, datetime.datetime):
        # Default XML-RPC handler is configured to decode a dateTime.iso8601 value
        # to builtin datetime.datetime instance
        return date
    if isinstance(date, xmlrpc.client.DateTime):
        # If a special xmlrpc.client.DateTime instance is given, convert it to a standard datetime instance
        return datetime.datetime.strptime(date.value, "%Y%m%dT%H:%M:%S")

    # If the date is given as str. This is the normal behavior for JSON-RPC
    try:
        return datetime.datetime.strptime(date, date_format)
    except ValueError:
        if raise_exception:
            raise
        return None


def ensure_sequence(element: Any) -> Sequence:
    """Ensure the given argument is a sequence object (tuple, list). If not, return a list containing its value."""
    return element if isinstance(element, (list, tuple)) else [element]


def first(seq: Iterable, default=NOT_SET) -> Any:
    """
    Return the first element of the given iterable, or default value if the iterable is empty.
    When the default value is not set, raise an IndexError if the iterable is empty.

    This helper should be used when RUF015 is encountered on linting.
    See https://docs.astral.sh/ruff/rules/unnecessary-iterable-allocation-for-first-element/
    """
    try:
        return next(iter(seq))
    except StopIteration:
        if default is not NOT_SET:
            return default
        raise IndexError("No element found in iterable") from None


def first_true(iterable: Iterable[Any], default: Any = None, pred: Callable[[Any], bool] | None = None) -> Any:
    """
    Same as more_itertools.first_true(), but avoid dependency to a new lib

    Doc: https://more-itertools.readthedocs.io/en/stable/api.html#more_itertools.first_true
    """
    return next(filter(pred, iterable), default)



from __future__ import annotations

import logging
from http import HTTPStatus
from typing import TYPE_CHECKING

from django.http.response import HttpResponse

from modernrpc.core import RpcRequestContext

if TYPE_CHECKING:
    from django.http import HttpRequest

    from modernrpc.server import RpcServer


logger = logging.getLogger(__name__)


def handle_rpc_request(request: HttpRequest, server: RpcServer) -> HttpResponse:
    """
    Synchronous view function to handle RPC requests.

    :param request: The HTTP request object
    :param server: The RPC server instance
    :return: An HTTP response object
    """
    response = server.check_request(request)
    if response:
        return response

    handler = server.get_request_handler(request)
    if not handler:
        return HttpResponse(
            f"Unable to handle your request, unsupported Content-Type {request.content_type}.",
            status=HTTPStatus.BAD_REQUEST,
            content_type=request.content_type,
        )

    result_data = handler.process_request(
        request.body.decode(request.encoding or server.default_encoding),
        RpcRequestContext(request, server, handler, handler.protocol),
    )

    return server.build_response(handler, result_data)


async def handle_rpc_request_async(request: HttpRequest, server: RpcServer) -> HttpResponse:
    """
    Asynchronous view function to handle RPC requests.

    :param request: The HTTP request object
    :param server: The RPC server instance
    :return: An HTTP response object
    """
    response = server.check_request(request)
    if response:
        return response

    handler = server.get_request_handler(request)
    if not handler:
        return HttpResponse(
            f"Unable to handle your request, unsupported Content-Type {request.content_type}.",
            status=HTTPStatus.BAD_REQUEST,
            content_type=request.content_type,
        )

    result_data = await handler.aprocess_request(
        request.body.decode(request.encoding or server.default_encoding),
        RpcRequestContext(request, server, handler, handler.protocol),
    )

    return server.build_response(handler, result_data)



# PEP 585: use of list[Any] instead of List[Any] is available since Python 3.9, enable it for older versions
# PEP 604: use of typeA | typeB is available since Python 3.10, enable it for older versions
from __future__ import annotations

from typing import TYPE_CHECKING, Iterable, overload

from modernrpc.constants import NOT_SET
from modernrpc.exceptions import RPCInvalidRequest
from modernrpc.jsonrpc.handler import JsonRpcRequest
from modernrpc.types import DictStrAny, RpcErrorResult

try:
    # types.NoneType is available only with Python 3.10+
    from types import NoneType
except ImportError:
    NoneType = type(None)  # type: ignore[misc]

if TYPE_CHECKING:
    from modernrpc.jsonrpc.handler import JsonRpcResult


class Unmarshaller:
    def __init__(self, validate_version: bool = True):
        self.validate_version = validate_version

    def validate_dict_request(self, request_data: DictStrAny) -> None:
        if "method" not in request_data:
            raise RPCInvalidRequest("Unable to find method name", data=request_data)

        if "jsonrpc" not in request_data:
            raise RPCInvalidRequest("jsonrpc required")

        if self.validate_version and request_data["jsonrpc"] != "2.0":
            raise RPCInvalidRequest(
                f'Parameter "jsonrpc" has an unsupported value "{request_data["jsonrpc"]}". It must be set to "2.0"'
            )

        # Notification request won't have "id" field
        # None is also an allowed value. Both cases are valid
        request_id = request_data.get("id")
        if type(request_id) not in (NoneType, int, float, str):
            raise RPCInvalidRequest(
                'Parameter "id" has an unsupported value. According to JSON-RPC 2.0 standard, it must '
                f"be a String, a Number or a Null value. Found: {type(request_id)}"
            )

    @overload
    def dict_to_request(self, structured_data: DictStrAny) -> JsonRpcRequest: ...

    @overload
    def dict_to_request(self, structured_data: list[DictStrAny]) -> list[JsonRpcRequest]: ...

    def dict_to_request(self, structured_data: DictStrAny | list[DictStrAny]) -> JsonRpcRequest | list[JsonRpcRequest]:
        if isinstance(structured_data, list):
            return [self.dict_to_request(data) for data in structured_data]

        self.validate_dict_request(structured_data)

        method_name = structured_data["method"]
        params = structured_data.get("params")
        args = params if isinstance(params, (list, tuple)) else []
        kwargs = params if isinstance(params, dict) else {}

        # Build request object. If "id" not in data, it's a notification request: request_id == NOT_SET
        return JsonRpcRequest(
            request_id=structured_data.get("id", NOT_SET),
            method_name=method_name,
            args=args,
            kwargs=kwargs,
        )


class Marshaller:
    @overload
    def result_to_dict(self, result: JsonRpcResult) -> DictStrAny | None: ...

    @overload
    def result_to_dict(self, result: Iterable[JsonRpcResult]) -> list[DictStrAny | None]: ...

    def result_to_dict(
        self, result: JsonRpcResult | Iterable[JsonRpcResult]
    ) -> DictStrAny | None | list[DictStrAny | None]:
        if isinstance(result, Iterable):
            return [self.result_to_dict(r) for r in result]

        if result.request.is_notification:
            return None

        base_result = {
            "id": result.request.request_id,
            "jsonrpc": result.request.jsonrpc,
        }

        if isinstance(result, RpcErrorResult):
            response_payload: DictStrAny = {
                **base_result,
                "error": {
                    "code": result.code,
                    "message": result.message,
                },
            }
            if result.data:
                response_payload["error"]["data"] = result.data
            return response_payload

        return {
            **base_result,
            "result": result.data,
        }



# PEP 585: use of list[Any] instead of List[Any] is available since Python 3.9, enable it for older versions
# PEP 604: use of typeA | typeB is available since Python 3.10, enable it for older versions
from __future__ import annotations

from typing import TYPE_CHECKING, Protocol

if TYPE_CHECKING:
    from modernrpc.xmlrpc.handler import XmlRpcRequest, XmlRpcResult


class XmlRpcDeserializer(Protocol):
    def __init__(self, **kwargs): ...

    def loads(self, data: str) -> XmlRpcRequest: ...


class XmlRpcSerializer(Protocol):
    def __init__(self, **kwargs): ...

    def dumps(self, result: XmlRpcResult) -> str: ...



# PEP 585: use of list[Any] instead of List[Any] is available since Python 3.9, enable it for older versions
# PEP 604: use of typeA | typeB is available since Python 3.10, enable it for older versions
from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import TYPE_CHECKING, Union

from django.utils.module_loading import import_string

from modernrpc import Protocol
from modernrpc.config import settings
from modernrpc.exceptions import RPCException
from modernrpc.handler import RpcHandler
from modernrpc.types import DictStrAny, RpcErrorResult, RpcRequest, RpcSuccessResult

if TYPE_CHECKING:
    from typing import ClassVar

    from modernrpc import RpcRequestContext
    from modernrpc.xmlrpc.backends import XmlRpcDeserializer, XmlRpcSerializer

logger = logging.getLogger(__name__)


@dataclass
class XmlRpcRequest(RpcRequest): ...


XmlRpcSuccessResult = RpcSuccessResult[XmlRpcRequest]
XmlRpcErrorResult = RpcErrorResult[XmlRpcRequest]
XmlRpcResult = Union[XmlRpcSuccessResult, XmlRpcErrorResult]


class XmlRpcHandler(RpcHandler[XmlRpcRequest]):
    """Default XML-RPC handler implementation"""

    protocol = Protocol.XML_RPC
    valid_content_types: ClassVar[list[str]] = ["text/xml", "application/xml"]
    response_content_type = "application/xml"
    success_result_type = XmlRpcSuccessResult
    error_result_type = XmlRpcErrorResult

    def __init__(self) -> None:
        deserializer_config = settings.MODERNRPC_XML_DESERIALIZER
        deserializer_klass: type[XmlRpcDeserializer] = import_string(deserializer_config["class"])
        deserializer_kwargs: DictStrAny = deserializer_config.get("kwargs", {})
        self.deserializer: XmlRpcDeserializer = deserializer_klass(**deserializer_kwargs)

        serializer_config = settings.MODERNRPC_XML_SERIALIZER
        serializer_klass: type[XmlRpcSerializer] = import_string(serializer_config["class"])
        serializer_kwargs: DictStrAny = serializer_config.get("kwargs", {})
        self.serializer: XmlRpcSerializer = serializer_klass(**serializer_kwargs)

    def process_request(self, request_body: str, context: RpcRequestContext) -> str:
        """
        Parse request and delegates to process_single_request(), catching exceptions to handle errors.

        `system.multicall()` is implemented in the `modernrpc.system_procedures` module.
        """
        try:
            request = self.deserializer.loads(request_body)

        except RPCException as exc:
            exc = context.server.on_error(exc, context)
            return self.serializer.dumps(self.build_error_result(XmlRpcRequest(method_name=""), exc.code, exc.message))

        result = self.process_single_request(request, context)

        try:
            return self.serializer.dumps(result)

        except RPCException as exc:
            exc = context.server.on_error(exc, context)
            return self.serializer.dumps(self.build_error_result(request, exc.code, exc.message))

    async def aprocess_request(self, request_body: str, context: RpcRequestContext) -> str:
        """
        Parse request and delegates to process_single_request(), catching exceptions to handle errors.

        `system.multicall()` is implemented in the `modernrpc.system_procedures` module.
        """
        try:
            request = self.deserializer.loads(request_body)

        except RPCException as exc:
            exc = context.server.on_error(exc, context)
            return self.serializer.dumps(self.build_error_result(XmlRpcRequest(method_name=""), exc.code, exc.message))

        result = await self.aprocess_single_request(request, context)

        try:
            return self.serializer.dumps(result)

        except RPCException as exc:
            exc = context.server.on_error(exc, context)
            return self.serializer.dumps(self.build_error_result(request, exc.code, exc.message))



# PEP 585: use of list[Any] instead of List[Any] is available since Python 3.9, enable it for older versions
# PEP 604: use of typeA | typeB is available since Python 3.10, enable it for older versions
from __future__ import annotations

import asyncio
import logging
from dataclasses import dataclass, field
from http import HTTPStatus
from typing import TYPE_CHECKING, Union

from django.utils.module_loading import import_string

from modernrpc import Protocol
from modernrpc.config import settings
from modernrpc.constants import NOT_SET
from modernrpc.exceptions import RPCException
from modernrpc.handler import RpcHandler
from modernrpc.types import DictStrAny, RpcErrorResult, RpcRequest, RpcSuccessResult

if TYPE_CHECKING:
    from typing import ClassVar, Generator

    from modernrpc import RpcRequestContext
    from modernrpc.jsonrpc.backends import JsonRpcDeserializer, JsonRpcSerializer

logger = logging.getLogger(__name__)


@dataclass
class JsonRpcRequest(RpcRequest):
    """
    JSON-RPC request specific data
    """

    kwargs: DictStrAny = field(default_factory=dict)
    request_id: RequestIdType | object = NOT_SET
    jsonrpc: str = "2.0"

    @property
    def is_notification(self) -> bool:
        return self.request_id is NOT_SET


RequestIdType = Union[str, int, float, None]
JsonRpcSuccessResult = RpcSuccessResult[JsonRpcRequest]
JsonRpcErrorResult = RpcErrorResult[JsonRpcRequest]
JsonRpcResult = Union[JsonRpcSuccessResult, JsonRpcErrorResult]


class JsonRpcHandler(RpcHandler[JsonRpcRequest]):
    """Default JSON-RPC handler implementation"""

    protocol = Protocol.JSON_RPC
    valid_content_types: ClassVar[list[str]] = ["application/json", "application/json-rpc", "application/jsonrequest"]
    response_content_type = "application/json"
    success_result_type = JsonRpcSuccessResult
    error_result_type = JsonRpcErrorResult

    def __init__(self) -> None:
        deserializer_config = settings.MODERNRPC_JSON_DESERIALIZER
        deserializer_klass: type[JsonRpcDeserializer] = import_string(deserializer_config["class"])
        deserializer_kwargs: DictStrAny = deserializer_config.get("kwargs", {})
        self.deserializer: JsonRpcDeserializer = deserializer_klass(**deserializer_kwargs)

        serializer_config = settings.MODERNRPC_JSON_SERIALIZER
        serializer_klass: type[JsonRpcSerializer] = import_string(serializer_config["class"])
        serializer_kwargs: DictStrAny = serializer_config.get("kwargs", {})
        self.serializer: JsonRpcSerializer = serializer_klass(**serializer_kwargs)

    def process_request(self, request_body: str, context: RpcRequestContext) -> str | tuple[HTTPStatus, str]:
        """
        Parse request and process it, according to its kind. Standard request as well as batch request is supported.

        Delegates to `process_single_request()` to perform most of the checks and procedure execution. Depending on the
        result of `parse_request()`, a standard or a batch request will be handled here.
        """
        try:
            parsed_request = self.deserializer.loads(request_body)

        except RPCException as exc:
            # We can't extract request_id from an incoming request. According to the spec, a
            # null 'id' should be used in response payload
            fake_request = JsonRpcRequest(request_id=None, method_name="")
            exc = context.server.on_error(exc, context)
            return self.serializer.dumps(self.build_error_result(fake_request, exc.code, exc.message))

        # Parsed request is an Iterable, we should handle it as a batch request
        if isinstance(parsed_request, list):
            return self.process_batch_request(parsed_request, context)

        # By default, handle a standard single request
        result = self.process_single_request(parsed_request, context)

        if parsed_request.is_notification:
            return HTTPStatus.NO_CONTENT, ""

        try:
            return self.serializer.dumps(result)
        except RPCException as exc:
            exc = context.server.on_error(exc, context)
            return self.serializer.dumps(self.build_error_result(parsed_request, exc.code, exc.message))

    async def aprocess_request(self, request_body: str, context: RpcRequestContext) -> str | tuple[HTTPStatus, str]:
        """
        Parse request and process it, according to its kind. Standard request as well as batch request is supported.

        Delegates to `process_single_request()` to perform most of the checks and procedure execution. Depending on the
        result of `parse_request()`, a standard or a batch request will be handled here.
        """
        try:
            parsed_request = self.deserializer.loads(request_body)

        except RPCException as exc:
            # We can't extract request_id from an incoming request. According to the spec, a
            # null 'id' should be used in response payload
            fake_request = JsonRpcRequest(request_id=None, method_name="")
            exc = context.server.on_error(exc, context)
            return self.serializer.dumps(self.build_error_result(fake_request, exc.code, exc.message))

        # Parsed request is an Iterable, we should handle it as a batch request
        if isinstance(parsed_request, list):
            return await self.aprocess_batch_request(parsed_request, context)

        # By default, handle a standard single request
        result = await self.aprocess_single_request(parsed_request, context)

        if parsed_request.is_notification:
            return HTTPStatus.NO_CONTENT, ""

        try:
            return self.serializer.dumps(result)
        except RPCException as exc:
            exc = context.server.on_error(exc, context)
            return self.serializer.dumps(self.build_error_result(parsed_request, exc.code, exc.message))

    def process_batch_request(
        self, requests: list[JsonRpcRequest], context: RpcRequestContext
    ) -> str | tuple[HTTPStatus, str]:
        # Process each request and store corresponding results (success or error)
        results: Generator[JsonRpcResult] = (self.process_single_request(request, context) for request in requests)

        # Filter out notification results
        filtered_results = [result for result in results if not result.request.is_notification]

        # Return JSON-serialized response list
        if filtered_results:
            return self.serializer.dumps(filtered_results)

        # Notifications-only batch request returns 204 no content
        return HTTPStatus.NO_CONTENT, ""

    async def aprocess_batch_request(
        self, requests: list[JsonRpcRequest], context: RpcRequestContext
    ) -> str | tuple[HTTPStatus, str]:
        # Process each request and store corresponding results (success or error)
        results: list[JsonRpcResult] = await asyncio.gather(
            *(self.aprocess_single_request(request, context) for request in requests)
        )

        # Filter out notification results
        filtered_results = [result for result in results if not result.request.is_notification]

        # Return JSON-serialized response list
        if filtered_results:
            return self.serializer.dumps(filtered_results)

        # Notifications-only batch request returns 204 no content
        return HTTPStatus.NO_CONTENT, ""



# PEP 585: use of list[Any] instead of List[Any] is available since Python 3.9, enable it for older versions
# PEP 604: use of typeA | typeB is available since Python 3.10, enable it for older versions
from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING, Iterable

import orjson
from django.utils.module_loading import import_string

from modernrpc.exceptions import RPCMarshallingError, RPCParseError

if TYPE_CHECKING:
    from modernrpc.jsonrpc.handler import JsonRpcRequest, JsonRpcResult
    from modernrpc.types import CustomKwargs, DictStrAny


class OrjsonDeserializer:
    """json-rpc deserializer based on the third-party orjson library"""

    def __init__(
        self,
        unmarshaller_klass="modernrpc.jsonrpc.backends.marshalling.Unmarshaller",
        unmarshaller_kwargs: CustomKwargs = None,
    ):
        self.unmarshaller_klass = import_string(unmarshaller_klass)
        self.unmarshaller_kwargs = unmarshaller_kwargs or {}
        # Info: orjson.loads() has not additional argument. Passing  load_kwargs to this class won't do anything

    @cached_property
    def unmarshaller(self):
        return self.unmarshaller_klass(**self.unmarshaller_kwargs)

    def loads(self, data: str) -> JsonRpcRequest | list[JsonRpcRequest]:
        try:
            structured_data: list[DictStrAny] | DictStrAny = orjson.loads(data)
        except orjson.JSONDecodeError as exc:
            raise RPCParseError(exc.msg, data=exc) from exc

        return self.unmarshaller.dict_to_request(structured_data)


class OrjsonSerializer:
    """json-rpc serializer based on the third-party orjson library"""

    def __init__(
        self,
        marshaller_klass="modernrpc.jsonrpc.backends.marshalling.Marshaller",
        marshaller_kwargs: CustomKwargs = None,
        dump_kwargs: CustomKwargs = None,
    ):
        self.marshaller_klass = import_string(marshaller_klass)
        self.marshaller_kwargs = marshaller_kwargs or {}

        self.dump_kwargs = dump_kwargs or {}

    @cached_property
    def marshaller(self):
        return self.marshaller_klass(**self.marshaller_kwargs)

    def dumps(self, result: JsonRpcResult | Iterable[JsonRpcResult]) -> str:
        structured_data = self.marshaller.result_to_dict(result)
        try:
            return orjson.dumps(structured_data, **self.dump_kwargs).decode("utf-8")
        except (orjson.JSONEncodeError, UnicodeDecodeError, TypeError) as exc:
            raise RPCMarshallingError(structured_data, exc) from exc



from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING

import defusedxml
import defusedxml.ElementTree as DefusedElementTree
from django.utils.module_loading import import_string

from modernrpc.exceptions import RPCInsecureRequest, RPCInvalidRequest, RPCMarshallingError, RPCParseError

if TYPE_CHECKING:
    from xml.etree.ElementTree import Element

    from modernrpc.types import CustomKwargs
    from modernrpc.xmlrpc.handler import XmlRpcRequest, XmlRpcResult


ETREE_DEFAULT_ELEMENT_TYPE = "xml.etree.ElementTree.Element"


class EtreeDeserializer:
    """xml-rpc deserializer based on python builtin module xml.etree"""

    def __init__(
        self,
        unmarshaller_klass="modernrpc.xmlrpc.backends.marshalling.EtreeElementUnmarshaller",
        unmarshaller_kwargs: CustomKwargs = None,
        element_type_klass=ETREE_DEFAULT_ELEMENT_TYPE,
        load_kwargs: CustomKwargs = None,
    ):
        self.unmarshaller_klass = import_string(unmarshaller_klass)
        self.unmarshaller_kwargs = unmarshaller_kwargs or {}

        self.element_type_klass = import_string(element_type_klass)

        self.load_kwargs = load_kwargs or {}
        self.load_kwargs.setdefault("forbid_dtd", True)

    @cached_property
    def unmarshaller(self):
        return self.unmarshaller_klass[self.element_type_klass](**self.unmarshaller_kwargs)

    def loads(self, data: str) -> XmlRpcRequest:
        try:
            root_obj: Element = DefusedElementTree.XML(data, **self.load_kwargs)
        except DefusedElementTree.ParseError as exc:
            raise RPCParseError(str(exc)) from exc
        except (defusedxml.DTDForbidden, defusedxml.EntitiesForbidden, defusedxml.ExternalReferenceForbidden) as exc:
            raise RPCInsecureRequest(str(exc)) from exc

        try:
            return self.unmarshaller.element_to_request(root_obj)
        except Exception as exc:
            raise RPCInvalidRequest(str(exc)) from exc


class EtreeSerializer:
    """xml-rpc serializer based on python builtin module xml.etree"""

    def __init__(
        self,
        marshaller_klass="modernrpc.xmlrpc.backends.marshalling.EtreeElementMarshaller",
        marshaller_kwargs: CustomKwargs = None,
        element_type_klass=ETREE_DEFAULT_ELEMENT_TYPE,
        dump_kwargs: CustomKwargs = None,
    ):
        self.marshaller_klass = import_string(marshaller_klass)
        self.marshaller_kwargs = marshaller_kwargs or {}
        self.marshaller_kwargs.setdefault("element_factory", "xml.etree.ElementTree.Element")
        self.marshaller_kwargs.setdefault("sub_element_factory", "xml.etree.ElementTree.SubElement")

        self.element_type_klass = import_string(element_type_klass)

        self.dump_kwargs = dump_kwargs or {}

    @cached_property
    def marshaller(self):
        elt_factory = import_string(self.marshaller_kwargs.pop("element_factory"))
        sub_elt_factory = import_string(self.marshaller_kwargs.pop("sub_element_factory"))
        return self.marshaller_klass[self.element_type_klass](elt_factory, sub_elt_factory, **self.marshaller_kwargs)

    def dumps(self, result: XmlRpcResult) -> str:
        """Serialize an XmlRpcResult to an XML string."""
        try:
            root = self.marshaller.result_to_element(result)
        except Exception as exc:
            raise RPCMarshallingError(result.data, exc) from exc

        return DefusedElementTree.tostring(root, encoding="unicode", **self.dump_kwargs)



"""Handle introspection features for remote procedures and other callables:
- Help text
- Signature
"""

from __future__ import annotations

import inspect
import re
import typing
from typing import Callable

from django.utils.functional import cached_property


class Introspector:
    """Helper class to extract the signature of a callable and the type hint of its arguments & returns"""

    def __init__(self, function: Callable):
        self.func = function

    @cached_property
    def signature(self) -> inspect.Signature:
        return inspect.signature(self.func)

    @cached_property
    def _func_type_hints(self) -> dict[str, type]:
        return typing.get_type_hints(self.func)

    def get_arg_type_hint(self, arg: str) -> type | None:
        return self._func_type_hints.get(arg)

    def get_return_type_hint(self) -> type | None:
        return self._func_type_hints.get("return")


class DocstringParser:
    """Helper class to parse the docstring of a callable and extract arguments and return docs & types."""

    MULTIPLE_SPACES = re.compile(r"\s+")

    PARAM_REXP = re.compile(r"^[:@]param (\w+): ?(.*(?:\n[^:@].+)*)$", flags=re.MULTILINE)
    RETURN_REXP = re.compile(r"^[:@]return: ?(.*(?:\n[^:@].+)*)$", flags=re.MULTILINE)

    PARAM_TYPE_REXP = re.compile(r"^[:@]type (\w+): ?(.*)$", flags=re.MULTILINE)
    RETURN_TYPE_REXP = re.compile(r"^[:@]rtype ?: ?(.*)$", flags=re.MULTILINE)

    def __init__(self, function: Callable):
        self.func = function

    @staticmethod
    def _cleanup_spaces(subject: str) -> str:
        """Basically, replaces any number of consecutive space chars (space, newline, tab, etc.) to a single space"""
        return DocstringParser.MULTIPLE_SPACES.sub(" ", subject)

    @cached_property
    def _full_docstring(self) -> str:
        """Return the function docstring content returned by inspect.getdoc()"""
        return inspect.getdoc(self.func) or ""

    @cached_property
    def docstring(self) -> str:
        """Return the text part of the docstring block, excluding legacy typehints, parameters and return docs"""
        content = self._full_docstring
        for pattern in [self.PARAM_REXP, self.PARAM_TYPE_REXP, self.RETURN_REXP, self.RETURN_TYPE_REXP]:
            content = pattern.sub("", content)
        return content.strip()

    @cached_property
    def args_docstrings(self) -> dict[str, str]:
        """Return a dict with, for each argument, its name as key and documentation text as value. The dict
        will contain only documented arguments.
        Basically this method parses and extracts reST documented arguments:

            :param <argname>: Documentation for <argname>

        Alternatively, it also supports the @param format:

            @param <argname>: Documentation for <argname>
        """
        return {param: self._cleanup_spaces(text) for param, text in self.PARAM_REXP.findall(self._full_docstring)}

    @cached_property
    def args_types(self) -> dict[str, str]:
        """Return a dict with, for each argument, its name as key and documented type as value. The dict
        will contain only documented arguments.
        Basically this method parses and extracts reST doctype documentation:

            :type <argname>: int or str

        Alternatively, it also supports the @type format:

            @type <argname>: int or str
        """
        return dict(self.PARAM_TYPE_REXP.findall(self._full_docstring))

    @cached_property
    def return_doc(self) -> str:
        """Return the documentation for the return value, as found in the docstring:

            :return: Documentation on return value

        Alternatively, it also supports @return format:

            @return: Documentation on return value

        Return an empty string if return documentation is not found
        """
        res = self.RETURN_REXP.search(self._full_docstring)
        return self._cleanup_spaces(res.group(1)) if res else ""

    @cached_property
    def return_type(self) -> str:
        """Return the documentation for the return type, as found in the docstring:

            :rtype: list

        Alternatively, it also supports the @rtype format:

            @rtype: list

        Return an empty string if return type information is not found
        """
        res = self.RETURN_TYPE_REXP.search(self._full_docstring)
        return res.group(1) if res else ""
